@@ -1468,12 +1468,6 @@ rsa_decrypt (gcry_sexp_t *r_plain, gcry_sexp_t s_data, gcry_sexp_t keyparms)
       rc = GPG_ERR_INV_DATA;
       goto leave;
     }
-  if (fips_mode () && (ctx.encoding == PUBKEY_ENC_PKCS1 ||
-                       ctx.encoding == PUBKEY_ENC_OAEP))
-    {
-      rc = GPG_ERR_INV_FLAG;
-      goto leave;
-    }
 
   /* Extract the key.  */
   rc = sexp_extract_param (keyparms, NULL, "nedp?q?u?",
@@ -2191,10 +2185,13 @@ selftests_rsa (selftest_report_func_t report, int extended)
   if (errtxt)
     goto failed;
 
-  what = "encrypt";
-  errtxt = selftest_encr_2048 (pkey, skey);
-  if (errtxt)
-    goto failed;
+  if (extended)
+    {
+      what = "encrypt";
+      errtxt = selftest_encr_2048 (pkey, skey);
+      if (errtxt)
+        goto failed;
+    }
 
   sexp_release (pkey);
   sexp_release (skey);
